import shutil
from pathlib import Path

from .core import ContextDir, Step
from .dependencies import pip_dependency


class ResourcePackerStep(Step):
    encoding = 'utf-8'

    def __init__(self, source_dir: ContextDir = 'input_dir'):
        self.source_dir: ContextDir = source_dir

    def __call__(self, path: Path, output_paths: list[Path]):
        parent_dir = self.context[self.source_dir]
        data = '\n\n'.join(
            (parent_dir / f).read_text(self.encoding)
            for filename in path.open() if (f := filename.strip()) and not f.startswith('#')
        )

        for o_path in output_paths:
            o_path.parent.mkdir(parents=True, exist_ok=True)
        output_paths[0].write_text(data, self.encoding)
        for o_path in output_paths[1:]:
            shutil.copy(output_paths[0], o_path)


class CSSMinifierStep(Step):
    encoding = 'utf-8'

    @classmethod
    def get_dependencies(cls):
        return {
            pip_dependency('lightningcss')
        }

    def __init__(self,
                 error_recovery: bool = False,
                 parser_flags: dict[str, bool] | None = None,
                 unused_symbols: set[str] | None = None,
                 browsers_list: list[str] | None = ['defaults'],
                 minify: bool = True):

        self.error_recovery = error_recovery
        self.parser_flags = parser_flags or {}
        self.unused_symbols = unused_symbols
        self.browsers_list = browsers_list
        self.minify = minify

    def __call__(self, path: Path, output_paths: list[Path]):
        import lightningcss
        data = lightningcss.process_stylesheet(
            path.read_text(self.encoding),
            filename=str(path),
            error_recovery=self.error_recovery,
            parser_flags=lightningcss.calc_parser_flags(**self.parser_flags),
            unused_symbols=self.unused_symbols,
            browsers_list=self.browsers_list,
            minify=self.minify
        )
        for o_path in output_paths:
            o_path.parent.mkdir(parents=True, exist_ok=True)
        output_paths[0].write_text(data, self.encoding)
        for o_path in output_paths[1:]:
            shutil.copy(output_paths[0], o_path)


class HTMLMinifierStep(Step):
    encoding = 'utf-8'
    minify_css = False
    minify_js = False

    @classmethod
    def get_dependencies(cls):
        return {
            (
                pip_dependency('minify-html-onepass', check_name='minify_html_onepass')
                | pip_dependency('minify-html', check_name='minify_html')
            ),
        }

    def __call__(self, path: Path, output_paths: list[Path]):
        params = {'minify_css': self.minify_css, 'minify_js': self.minify_js}
        try:
            from minify_html_onepass import minify
        except ImportError:
            from minify_html import minify
            params |= {
                'do_not_minify_doctype': True,
                'ensure_spec_compliant_unquoted_attribute_values': True,
                'keep_spaces_between_attributes': True,
            }
        data = minify(path.read_text(self.encoding), **params)
        for o_path in output_paths:
            o_path.parent.mkdir(parents=True, exist_ok=True)
        output_paths[0].write_text(data, self.encoding)
        for o_path in output_paths[1:]:
            shutil.copy(output_paths[0], o_path)



from __future__ import annotations

import shutil
import typing as t
from functools import reduce
from pathlib import Path

from .core import Context, Step
from .dependencies import pip_dependency, Dependency

if t.TYPE_CHECKING:
    from jinja2 import Environment


MDProcessor = t.Callable[[str], str]


class JinjaRenderStep(Step):
    """
    Abstract base class for Steps using Jinja rendering.
    """
    encoding = 'utf-8'
    env: Environment

    @classmethod
    def get_dependencies(cls):
        return {
            pip_dependency('jinja2'),
        }

    def __init__(self,
                 env: Environment | None = None,
                 extra_globals: dict[str, t.Any] | None = None):
        self._temporary_env = env
        self._extra_globals = extra_globals

    def bind(self, context: Context):
        """
        Bind this Step to a specific context. Also initializes a Jinja
        environment if none is set up already.
        """
        super().bind(context)

        if self._temporary_env:
            self.env = self._temporary_env
        else:
            from jinja2 import Environment, FileSystemLoader, select_autoescape
            self.env = Environment(
                loader=FileSystemLoader(context['input_dir']),
                autoescape=select_autoescape()
            )
        if self._extra_globals:
            self.env.globals.update(self._extra_globals)

    def render_template(self, template_name: str, meta: dict[str, t.Any], output_paths: list[Path]):
        """
        Look up a Jinja template by name and render it with @meta as
        parameters, then save the result to each of the provided @output_paths.
        """
        if not output_paths:
            return

        template = self.env.get_template(template_name)
        for path in output_paths:
            path.parent.mkdir(parents=True, exist_ok=True)
        template.stream(**meta).dump(str(output_paths[0]), encoding=self.encoding)
        for path in output_paths[1:]:
            shutil.copy(output_paths[0], path)


class JinjaMarkdownStep(JinjaRenderStep):
    """
    A Step for rendering Markdown using Jinja templates. Parses according to
    CommonMark and Renders to HTML by default.
    """
    @classmethod
    def _build_markdownit(cls):
        import markdown_it
        processor = markdown_it.MarkdownIt()

        def convert(md_string: str) -> str:
            return processor.render(md_string)

        return convert

    @classmethod
    def _build_mistletoe(cls):
        import mistletoe
        processor = mistletoe.HTMLRenderer()

        def convert(md_string: str) -> str:
            return processor.render(mistletoe.Document(md_string))

        return convert

    @classmethod
    def _build_markdown(cls):
        import markdown
        processor = markdown.Markdown()

        def convert(md_string: str):
            return processor.convert(md_string)

        return convert

    @classmethod
    def _build_commonmark(cls):
        import commonmark
        parser = commonmark.Parser()
        renderer = commonmark.HtmlRenderer()

        def convert(md_string: str) -> str:
            return renderer.render(parser.parse(md_string))

        return convert

    @classmethod
    def get_options(cls):
        return [
            (pip_dependency('markdown-it-py', None, 'markdown_it'), cls._build_markdownit),
            (pip_dependency('mistletoe'), cls._build_mistletoe),
            (pip_dependency('markdown'), cls._build_markdown),
            (pip_dependency('commonmark'), cls._build_commonmark),
        ]

    @classmethod
    def get_dependencies(cls):
        deps = [option[0] for option in cls.get_options()]
        dep_set = {reduce(lambda x, y: x | y, deps)} if deps else set[Dependency]()

        return super().get_dependencies() | dep_set

    def __init__(self,
                 default_template: str | None = None,
                 md_processor: MDProcessor | None = None,
                 jinja_env: Environment | None = None,
                 jinja_globals: dict[str, t.Any] | None = None):
        super().__init__(jinja_env, jinja_globals)
        self.default_template = default_template
        self._md_processor = md_processor

    @property
    def md_processor(self):
        if not self._md_processor:
            for dep, factory in self.get_options():
                if dep.satisfied:
                    self._md_processor = factory()
                    break
            else:
                raise RuntimeError('Markdown processor could not be initialized!')
        return self._md_processor


    def __call__(self, path: Path, output_paths: list[Path]):
        meta, content = self.extract_metadata(path.read_text(self.encoding))
        meta |= {'rendered_markdown': self.md_processor(content.strip()).strip()}

        self.render_template(
            meta.get('template', self.default_template),
            meta,
            output_paths
        )

    def extract_metadata(self, text: str):
        """
        Read metadata from the front of a markdown-formatted text.
        """
        meta = {}
        lines = text.splitlines()

        i = 0
        for line in lines:
            if ':' not in line:
                break
            key, value = line.split(':', 1)
            if not key.isidentifier():
                break

            meta[key.strip()] = value.strip()
            i += 1

        return meta, '\n'.join(lines[i:])



from .core import Context, InputBuildSettings, Matcher, PathCalc, Rule, Step
from .css import AnchovyCSSStep
from .dependencies import (
    Dependency,
    import_install_check,
    pip_dependency,
    web_exec_dependency,
    which_install_check,
)
from .images import CWebPStep, ImageMagickStep, IMThumbnailStep, PillowStep, OptipngStep
from .jinja import JinjaMarkdownStep, JinjaRenderStep
from .minify import CSSMinifierStep, HTMLMinifierStep, ResourcePackerStep
from .paths import DirPathCalc, OutputDirPathCalc, REMatcher, WorkingDirPathCalc
from .simple import DirectCopyStep



from __future__ import annotations

import importlib
import shutil
import typing as t


# TODO: I'm not a huge fan of this system. Might be better to have a
# DependencyType class that uses __init_subclass__ to register its children.
DEPENDENCY_TYPES: dict[str, tuple[str, t.Callable[[], bool]]] = {
    'pip': (
        'pip install {source}',
        lambda: True
    ),
    'web': (
        '{source}',
        lambda: True
    ),
    'npm': (
        'npm install -g {source}',
        lambda: True
    ),
}


def import_install_check(dependency: Dependency):
    """
    An install checker which tries to import a Python module.
    """
    try:
        importlib.import_module(dependency.check_name)
    except ImportError:
        return False
    return True


def which_install_check(dependency: Dependency):
    """
    An install checker using `shutil.which()` to look for an executable.
    """
    return bool(shutil.which(dependency.check_name))


def pip_dependency(name: str, source: str | None = None, check_name: str | None = None):
    """
    A shortcut function for creating typical pip-based Dependencys.
    """
    return Dependency(name, 'pip', import_install_check, source, check_name)


def web_exec_dependency(name: str, source: str | None = None, check_name: str | None = None):
    """
    A shortcut function for creating typical Dependencys for general
    internet-sourced executables.
    """
    return Dependency(name, 'web', which_install_check, source, check_name)


class Dependency:
    """
    A class for tracking and evaluating dependencies.
    """
    def __init__(self,
                 name: str,
                 type: str,
                 install_check: t.Callable[[Dependency], bool],
                 source: str | None = None,
                 check_name: str | None = None):
        self.name = name
        self.type = type
        self.install_check = install_check
        self.source = source or name
        self.check_name = check_name or name

    @property
    def satisfied(self):
        """
        A bool indicating if this dependency is met.
        """
        return self.install_check(self)

    @property
    def needed(self):
        """
        A bool indicating if this dependency is needed on the current platform.
        """
        return DEPENDENCY_TYPES[self.type][1]()

    @property
    def install_hint(self):
        """
        A string giving help on how to install this dependency.
        """
        return DEPENDENCY_TYPES[self.type][0].format(name=self.name, source=self.source)

    def __repr__(self):
        return f'Dependency(name={self.name}, needed={self.needed}, satisfied={self.satisfied})'

    def __str__(self):
        return self.name

    def __or__(self, other: Dependency):
        return _OrDependency(self, other)

    def __and__(self, other: Dependency):
        return _AndDependency(self, other)


class _OrDependency(Dependency):
    def __init__(self, left: Dependency, right: Dependency):
        self.left = left
        self.right = right

    def __repr__(self):
        return f'{self.left} | {self.right}'

    def __str__(self):
        return repr(self)

    @property
    def satisfied(self):
        """
        A bool indicating whether either of these dependencies are met.
        """
        return self.left.satisfied or self.right.satisfied

    @property
    def needed(self):
        """
        A bool indicating whether either of these dependencies is needed.
        """
        return self.left.needed or self.right.needed

    @property
    def install_hint(self):
        """
        A string giving help on how to meet one of these dependencies, if
        either is needed.
        """
        return (
            (self.left.needed and self.left.install_hint)
            or (self.right.needed and self.right.install_hint)
            or ''
        )


class _AndDependency(Dependency):
    def __init__(self, left: Dependency, right: Dependency):
        self.left = left
        self.right = right

    def __repr__(self):
        return f'{self.left} & {self.right}'

    def __str__(self):
        return repr(self)

    @property
    def satisfied(self):
        """
        A bool indicating whether both of these dependencies are met.
        """
        return self.left.satisfied and self.right.satisfied

    @property
    def needed(self):
        """
        A bool indicating whether either of these dependencies is needed.
        """
        # These should mostly match, but we'll allow for cases where one
        # dependency is paired on more specific platforms.
        return self.left.needed or self.right.needed

    @property
    def install_hint(self):
        """
        A string giving help on how to install these dependencies, if they are
        needed.
        """
        return '; '.join(d.install_hint for d in [self.left, self.right] if d.needed)
